import queue
import logging
from threading import Thread

import zmq.error

from .comm import ZmqAsyncConn

log = logging.getLogger(__name__)


class MsgReg:

    def __init__(self, i):
        self._q = queue.Queue()
        self._id = i

    def id_match(self, i: str):
        return self._id == i[0:len(self._id)]

    def add_msg(self, msg):
        self._q.put(msg)

    def get_msg(self):
        return self._q.get(block=True)


class DataManager:

    def __init__(self, addr):
        self._cdm_async = ZmqAsyncConn(addr)

        self._thread = Thread(target=self._frame_reader)
        self._thread.start()

        self._msg_reg = []

    def close(self):
        log.info("Closing DataManager")
        self._cdm_async.disconnect()
        self._thread.join()

    def register(self, s):
        reg = MsgReg(s)
        self._msg_reg.append(reg)
        return reg

    def _frame_reader(self):
        log.info("FRAME READER STARTED")
        try:
            while True:
                msg_id = self._cdm_async.recv().decode()
                msg = self._cdm_async.recv()
                for i in self._msg_reg:
                    if i.id_match(msg_id):
                        i.add_msg(msg)
        except zmq.error.ContextTerminated:
            log.warning("Exiting from frame reader")




import signal

from .cfg import Cfg
from .utils import check_valid_board_identifier
from .data_mgr import DataManager
from .cmd_mgr import CmdManager


class CdmCmd:
    def __init__(self, cmd):
        self._cmd = cmd

    def get_info(self):
        return self._cmd.cdm_get_info()

    def get_sm_state(self):
        return self._cmd.cdm_get_state()

    def get_config(self):
        return self._cmd.get_cdm_config()

    def download_file(self, year, month, day, hour, fname, dpath):
        return self._cmd.download_file(year, month, day, hour, fname, dpath)

    def power_on(self, s=""):
        return self._cmd.power_on_ps(s)

    def power_off(self, s=""):
        return self._cmd.power_off_ps(s)

    def get_warnings(self):
        return self._cmd.get_warnings()


class BcgCmd:
    def __init__(self, cmd):
        self._cmd = cmd

    def get_running_partition(self):
        return self._cmd.get_running_partition()

    def get_info(self):
        return self._cmd.get_info()

    def get_voltage(self):
        return self._cmd.get_voltage()

    def get_adc_map(self):
        return self._cmd.get_adc_map()

    def get_config(self):
        return self._cmd.get_configuration()

    def get_state(self):
        return self._cmd.get_state()

    def dump_log(self, idns: list[str]):
        """
        :param idns: List of valid identifiers. Can be any of {R,L} or {R,L}{0,1,2}
        """
        for i in idns:
            check_valid_board_identifier(i)

        return self._cmd.dump_log(idns)

    def hard_reset(self):
        return self._cmd.hard_reset()


class EbisCmd:
    def __init__(self, cmd):
        self._cmd = cmd

    # XXX: Maybe this function should be moved to CDM class as it's the CDM who is connecting to the EBI
    # the EBI is not connecting anywhere
    def connect(self):
        self._cmd.connect()
        return self.version()

    def disconnect(self):
        return self._cmd.disconnect()

    def echo(self, ebi):
        return self._cmd.ebi_echo(ebi)

    def version(self):
        return self._cmd.baffle_get_version()

    def flash_ota(self, file):
        return self._cmd.flash_ebi_ota(file)

    def boot_ota(self):
        return self._cmd.boot_ebi_ota()

    def mark_ota_as_ok(self):
        return self._cmd.mark_ebi_ota_as_ok()


class BsdasCmd:
    def __init__(self, cmd):
        self._cmd = cmd

    def echo(self, ebi: str, bsda: int):
        return self._cmd.bsda_echo(ebi, bsda)

    def boot_ota(self):
        return self._cmd.boot_ota()

    def crash_system(self, idn: str):
        """
        :param idn: A board identifier, it must be one of {R,L}{0,1,2}
        """
        return self._cmd.crash_system(idn)

    def get_coredump(self, idn: str, out_path):
        """
        :param idn: A board identifier, it must be one of {R,L}{0,1,2}
        :param out_path: File where coredump will be saved
        """
        return self._cmd.get_coredump(idn, out_path)

    def prepare_ota(self):
        return self._cmd.prepare_ota()

    def flash_ota(self, file):
        return self._cmd.flash_ota(file)

    def mark_ota_as_ok(self):
        return self._cmd.mark_ota_as_ok()


class PyCDM:
    """
    This class exports only the bare minimum commands needed for the normal operation of the system.
    All other commands are exported by the different classes contained by this one.
    """
    def __init__(self, cdm_addr = ""):
        self._cfg = Cfg()
        if cdm_addr == "":
            cdm_addr = self._cfg.get_cdm_addr()
        self._cmd = CmdManager(cdm_addr)
        self._data = DataManager(cdm_addr)
        signal.signal(signal.SIGINT, self._handler)

        # Commands targeted to CDM
        self.cdm = CdmCmd(self._cmd)

        # Commands targeted to one or more BCG(s)
        self.bcg = BcgCmd(self._cmd)

        # Commands targeted at one or more EBI(s)
        self.ebi = EbisCmd(self._cmd)

        # Commands targeted at one or more BSDA(s)
        self.bsda = BsdasCmd(self._cmd)

    def _handler(self, signum, frame):
        self.close()

    def close(self):
        self._data.close()

    def check_comm(self):
        return self._cmd.check_comm()

    def configure(self, baffle_settings, right_baffle_settings=None):
        return self._cmd.configure(baffle_settings, right_baffle_settings)

    def start_acq(self):
        return self._cmd.start_acq()

    def stop_acq(self):
        return self._cmd.stop_acq()

    def read_pt100(self, ebi: str, bsda: int):
        uv = self._cmd.read_pt100(ebi, bsda)["ans"]
        uva = uv[0]
        uvb = uv[1]

        r1 = 500
        vcc = 3.37
        vdiff = abs(uva-uvb)/1000.0/1000.0

        rpt100 = (r1 * vdiff + 100 * vcc) / (vcc - vdiff)
        return uv, r1, (rpt100 - 100.0) / 0.39083

    def reg_msg(self, s):
        return self._data.register(s)


import socket

import zmq

context = zmq.Context()


class ZmqSyncConn:
    def __init__(self, cdm_addr):
        t = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        t.settimeout(2)
        if t.connect_ex((cdm_addr, 32000)) != 0:
            raise Exception("CDM doesn't seems to be running")
        self._sock = context.socket(zmq.REQ)
        self._sock.connect("tcp://{}:32000".format(cdm_addr))

    def send(self, s):
        self._sock.send_string(s)

    def recv(self):
        return self._sock.recv()


class ZmqAsyncConn:
    def __init__(self, cdm_addr):
        self._sock = context.socket(zmq.SUB)
        self._ctxt = self._sock.connect("tcp://{}:32001".format(cdm_addr))
        self._sock.setsockopt_string(zmq.SUBSCRIBE, "")

    def recv(self):
        return self._sock.recv()

    def disconnect(self):
        return context.destroy()



import json
from typing import Any

import cdmpy

from .comm import ZmqSyncConn
from .structs import EbiSettings
from .utils import *

"""
for the list of implemented commands see cmd.cpp:CmdExecutor::execute
"""
class CmdManager:
    def __init__(self, addr):
        self._cdm_cmd = ZmqSyncConn(addr)

    def ebi_echo(self, ebi):
        return self._process_cmd(cdmpy.ebi_echo, {"ebi": ebi})

    def bsda_echo(self, ebi, bsda):
        return self._process_cmd(cdmpy.bsda_echo, {"ebi": ebi, "bsda": bsda})

    def baffle_get_version(self):
        return self._process_cmd(cdmpy.get_version)

    def cdm_get_info(self):
        return self._process_cmd(cdmpy.cdm_version)

    def cdm_get_state(self):
        return cdmpy.SM_STATE(self._process_cmd(cdmpy.get_cdm_state)["state"])

    def connect(self):
        return self._process_cmd(cdmpy.connect)

    def disconnect(self):
        return self._process_cmd(cdmpy.disconnect)

    def configure(self, baffle_settings, right_baffle_settings=None):
        if right_baffle_settings:
            d = {"left": baffle_settings.to_json(), "right": right_baffle_settings.to_json()}
            return self._process_cmd(cdmpy.configure, d)
        else:
            d = {"left": baffle_settings.to_json(), "right": baffle_settings.to_json()}
            return self._process_cmd(cdmpy.configure, d)

    def get_configuration(self):
        s = self._process_cmd(cdmpy.get_config)
        left = EbiSettings().from_json(s["left"])
        right = EbiSettings().from_json(s["right"])
        return left, right

    def start_acq(self):
        return self._process_cmd(cdmpy.start)

    def stop_acq(self):
        return self._process_cmd(cdmpy.stop)

    def check_comm(self):
        return self._process_cmd(cdmpy.check)

    def get_state(self):
        return self._process_cmd(cdmpy.get_state)

    def download_file(self, year, month, day, hour, fname, dpath):
        d = {"year": year, "month": month, "day": day, "hour": hour, "file_name": fname}
        ans = self._process_cmd(cdmpy.download_file, d)
        decode_to_file(ans["file_content"], dpath)
        del ans["file_content"]
        return ans

    def get_cdm_config(self):
        return self._process_cmd(cdmpy.get_cdm_config)

    def hard_reset(self, reset_bsda=False):
        return self._process_cmd(cdmpy.hard_reset, {"reset_bsdas": reset_bsda})

    def crash_system(self, idn: str):
        """
        :param idn: A board identifier, it must be one of {R,L}{0,1,2}
        """
        check_valid_board_identifier(idn)
        return self._process_cmd(cdmpy.crash, {"board_identifier": idn})

    def get_coredump(self, idn: str, out_path):
        """
        :param idn: A board identifier, it must be one of {R,L}{0,1,2}
        :param out_path: File where coredump will be saved
        """
        idn = check_valid_board_identifier(idn)
        # check that a BSDA is select and not only one of the ERBIs
        if len(idn) != 2:
            raise Exception("Invalid board identifier")

        ans = self._process_cmd(cdmpy.get_coredump, {"board_identifier": idn})
        decode_to_file(ans["coredump"], out_path)
        del ans["coredump"]
        return ans

    def dump_log(self, idns):
        for i in idns:
            check_valid_board_identifier(i)

        args = {"board_identifiers": idns}
        ans = self._process_cmd(cdmpy.dump_log, args)

        for k, v in ans.items():
            ans[k] = v.splitlines()

        return ans

    def prepare_ota(self):
        return self._process_cmd(cdmpy.prepare_bsda_ota)

    def flash_ota(self, fname):
        return self._flash(fname, cdmpy.flash_bsda_ota)

    def boot_ota(self):
        return self._process_cmd(cdmpy.boot_bsda_ota)

    def mark_ota_as_ok(self):
        return self._process_cmd(cdmpy.mark_bsda_ota_ok)

    def flash_ebi_ota(self, fname):
        return self._flash(fname, cdmpy.flash_ebi_ota)

    def boot_ebi_ota(self):
        return self._process_cmd(cdmpy.boot_ebi_ota)

    def mark_ebi_ota_as_ok(self):
        return self._process_cmd(cdmpy.mark_ebi_ota_ok)

    def get_running_partitions(self):
        return self._process_cmd(cdmpy.get_partition)

    def get_info(self):
        return self._process_cmd(cdmpy.get_info)

    def get_voltage(self):
        return self._process_cmd(cdmpy.get_voltage)

    def get_adc_map(self):
        return self._process_cmd(cdmpy.get_adc_map)

    def read_pt100(self, ebi: str, bsda: int):
        return self._process_cmd(cdmpy.read_pt100, {"ebi": ebi, "bsda": bsda})

    def power_on_ps(self, s):
        if s != "I_KNOW_WHAT_I_AM_DOING":
            raise Exception("Not allowed to use this method")

        return self._process_cmd(cdmpy.power_on, {"is_engineer": True})

    def power_off_ps(self, s):
        if s != "I_KNOW_WHAT_I_AM_DOING":
            raise Exception("Not allowed to use this method")

        return self._process_cmd(cdmpy.power_off, {"is_engineer": True})

    def get_warnings(self):
        return self._process_cmd(cdmpy.get_warnings)

    def _flash(self, fname, cmd):
        fw, h = encode_from_file(fname)
        args = {"hash": h, "firmware_data": fw.decode()}
        return self._process_cmd(cmd, args)

    def _process_cmd(self, cmd: str, args: dict[str, Any] = {}):
        msg = json.dumps({"cmd": cmd, "args": args})
        self._cdm_cmd.send(msg)

        ans = json.loads(self._cdm_cmd.recv())
        if ans is None:
            return {}

        err = ans.pop("error", None)
        if err is not None:
            raise Exception(err)

        return ans
